using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using FooCfgToXml;

namespace UiColumnsInterpreters.ItemInterpreters
{
	public class ButtonsItemInterpreter : IItemInterpreter
	{
		private static readonly Guid sGuid = new Guid("d8e65660-64ed-42e7-850b-31d828c25294");
		
		private static readonly string[] sApperances = new string[] { "Normal", "Flat", "No edges" };
		private static readonly string[] sCommandGroups = new string[] { "Separator", "Buttons", "Context menu items", "Main menu items" };
		private static readonly string[] sItemGroups = new string[] { "None", "Now playing item", "Current playlist selection" };
		private static readonly string[] sDisplayStyles = new string[] { "Image", "Image and text", "Text" };
		
		public Guid TypeGuid
		{
			get { return sGuid; }
		}
		public string TypeName
		{
			get { return "Buttons"; }
		}

		public void WriteValueToXml(XmlWriter writer, byte[] itemSettings, Dictionary<Guid, IItemInterpreter> itemInterpreters)
		{
			if (itemSettings.Length == 0)
				return;

			MemoryStream reader = new MemoryStream(itemSettings, false);

			writer.WriteStartElement("Buttons");
			writer.WriteAttributeString("flag", XmlConvert.ToString(CfgToXmlUtilities.ReadInt(reader))); //This always appears to be 1. Don't know what its for, so give it a generic name

			writer.WriteAttributeString("textLocation", TextLocationToString(CfgToXmlUtilities.ReadBool(reader)));
			writer.WriteAttributeString("appearance", CfgToXmlUtilities.ReadEnum(reader, sApperances));

			int numberOfButtons = CfgToXmlUtilities.ReadInt(reader);

			for (int i = 0; i < numberOfButtons; i++)
			{
				writer.WriteStartElement("Button");
				writer.WriteAttributeString("commandGroup", CfgToXmlUtilities.ReadEnum(reader, sCommandGroups));
				writer.WriteAttributeString("itemGroup", CfgToXmlUtilities.ReadEnum(reader, sItemGroups));

				//Read now, to write later
				byte[] commandIdentifierBytes = new byte[32];
				reader.Read(commandIdentifierBytes, 0, commandIdentifierBytes.Length);

				writer.WriteAttributeString("display", CfgToXmlUtilities.ReadEnum(reader, sDisplayStyles));

				writer.WriteStartElement("CommandIdentifier");
				CfgToXmlUtilities.WriteByteArrayToXml(writer, commandIdentifierBytes);
				writer.WriteEndElement(); //CommandIdentifer

				bool useCustomNormalImage = CfgToXmlUtilities.ReadBool(reader);
				if (useCustomNormalImage)
				{
					writer.WriteStartElement("NormalImage");
					WriteImageToXml(writer, reader);
					writer.WriteEndElement(); //NormalImage
				}
				bool useCustomHotImage = CfgToXmlUtilities.ReadBool(reader);
				if (useCustomHotImage)
				{
					writer.WriteStartElement("HotImage");
					WriteImageToXml(writer, reader);
					writer.WriteEndElement(); //HotImage
				}

				bool useCustomText = CfgToXmlUtilities.ReadBool(reader);
				if (useCustomText)
				{
					writer.WriteStartElement("Text");
					writer.WriteString(CfgToXmlUtilities.ReadString(reader, false));
					writer.WriteEndElement(); //Text
				}

				writer.WriteEndElement(); //Button
			}

			writer.WriteEndElement(); //Buttons
		}

		public void WriteXmlToValue(Stream writer, XmlElement itemSettings, Dictionary<string, IItemInterpreter> itemInterpreters)
		{
			if (itemSettings.ChildNodes.Count == 0)
				return;

			XmlElement buttons = (XmlElement)itemSettings.FirstChild; //Should be first and only child
			XmlToCfgUtilities.WriteInt(writer, XmlConvert.ToInt32(buttons.GetAttribute("flag")));

			XmlToCfgUtilities.WriteBool(writer, StringToTextLocation(buttons.GetAttribute("textLocation")));
			XmlToCfgUtilities.WriteEnum(writer, buttons.GetAttribute("appearance"), sApperances);

			XmlNodeList buttonElements = buttons.SelectNodes("Button");
			XmlToCfgUtilities.WriteInt(writer, buttonElements.Count);
			foreach (XmlElement buttonElement in buttonElements)
			{
				XmlToCfgUtilities.WriteEnum(writer, buttonElement.GetAttribute("commandGroup"), sCommandGroups);
				XmlToCfgUtilities.WriteEnum(writer, buttonElement.GetAttribute("itemGroup"), sItemGroups);

				XmlToCfgUtilities.WriteXmlToByteArray(writer, (XmlElement)buttonElement.SelectSingleNode("CommandIdentifier"));

				XmlToCfgUtilities.WriteEnum(writer, buttonElement.GetAttribute("display"), sDisplayStyles);

				XmlElement normalImage = (XmlElement)buttonElement.SelectSingleNode("NormalImage");
				if (normalImage == null)
				{
					XmlToCfgUtilities.WriteBool(writer, false);
				}
				else
				{
					XmlToCfgUtilities.WriteBool(writer, true);
					WriteXmlToImage(writer, normalImage);
				}
				XmlElement hotImage = (XmlElement)buttonElement.SelectSingleNode("HotImage");
				if (hotImage == null)
				{
					XmlToCfgUtilities.WriteBool(writer, false);
				}
				else
				{
					XmlToCfgUtilities.WriteBool(writer, true);
					WriteXmlToImage(writer, hotImage);
				}

				XmlElement customText = (XmlElement)buttonElement.SelectSingleNode("Text");
				if (customText == null)
				{
					XmlToCfgUtilities.WriteBool(writer, false);
				}
				else
				{
					XmlToCfgUtilities.WriteBool(writer, true);
					XmlToCfgUtilities.WriteString(writer, customText.InnerText, true, false);
				}
			}
		}

		private static void WriteImageToXml(XmlWriter writer, MemoryStream reader)
		{
			writer.WriteAttributeString("path", CfgToXmlUtilities.ReadString(reader, false));
			int maskType = CfgToXmlUtilities.ReadInt(reader);
			switch (maskType)
			{
				case 0: //None
					break;
				case 1: //Bitmap
					writer.WriteStartElement("MaskBitmap");
					writer.WriteAttributeString("path", CfgToXmlUtilities.ReadString(reader, false));
					writer.WriteEndElement(); //MaskImage
					break;
				case 2: //Color
					writer.WriteStartElement("MaskColor");
					Utilities.WriteColorToXml(writer, reader);
					writer.WriteEndElement(); //MaskColor
					break;
				default:
					Console.WriteLine("Unexpected mask setting! Check results carefully");
					break;
			}
		}
		private static void WriteXmlToImage(Stream writer, XmlElement imageElement)
		{
			XmlToCfgUtilities.WriteString(writer, imageElement.GetAttribute("path"), true, false);
			XmlElement maskElement = (XmlElement)imageElement.FirstChild;
			if (maskElement == null)
			{
				//No mask
				XmlToCfgUtilities.WriteInt(writer, 0);
			}
			else if (maskElement.Name == "MaskBitmap")
			{
				XmlToCfgUtilities.WriteInt(writer, 1);
				XmlToCfgUtilities.WriteString(writer, maskElement.GetAttribute("path"), true, false);
			}
			else if (maskElement.Name == "MaskColor")
			{
				XmlToCfgUtilities.WriteInt(writer, 1);
				Utilities.WriteColorToCfg(writer, maskElement);
			}
			else
			{
				Console.WriteLine("Unexpected mask setting! Check results carefully");
			}
		}

		private string TextLocationToString(bool textLocation)
		{
			return textLocation ? "Below" : "Right";
		}
		private bool StringToTextLocation(string textLocation)
		{
			return textLocation == "Below";
		}
	}
}
