using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;
using System.Globalization;

namespace FooCfgToXml
{
	/// <summary>
	/// This class contains helper methods for reading cfg binary and writing to xml
	/// They are used internally, and can also be used by value interpreter extensions
	/// </summary>
	public static class CfgToXmlUtilities
	{
		//Whitelist of characters that do not have a generally valid unicode category (like Control), but are valid in strings nonetheless
		private static char[] sWhitelist = new char[] { '\r', '\n' };
	
		/// <summary>
		/// Writes out a byte array guessing at strings where possible.
		/// What isn't detected as a string is put out as binhex.
		/// </summary>
		public static void WriteByteArrayToXml(XmlWriter writer, byte[] bytes)
		{
			Encoding stringEncoding = Encoding.UTF8; //Probably.
			int maxByteCountPerChar = stringEncoding.GetMaxByteCount(1);

			char[] potentialString = new char[0];
			int potentialStringIndex = 0;
			int targetPotentialStringLength = 0;
			bool forceLengthDescriptor = false;
			int writeIndex = 0;
			for (int i = 0; i < bytes.Length; i++)
			{
				//Detect strings
				char[] chars = stringEncoding.GetChars(bytes, i, Math.Min(maxByteCountPerChar, bytes.Length - i));
				if (chars.Length > 0)
				{
					//sanity check: find the number of bytes the first char should have, and reparse for exactly that number of bytes
					int collapsedByteCount = stringEncoding.GetByteCount(chars, 0, 1);
					chars = stringEncoding.GetChars(bytes, i, collapsedByteCount);
				}

				if (chars.Length > 0 && (UnicodeCategoryIsValid(Char.GetUnicodeCategory(chars[0])) || //If this is a valid character
										Array.IndexOf<char>(sWhitelist, chars[0]) >= 0 || //Some characters fail the category test, but are valid anyway
										chars[0] == 0 && potentialStringIndex == targetPotentialStringLength - 1)) //or is the null terminating character
				{
					int byteDelta = stringEncoding.GetByteCount(chars, 0, 1) - 1; //If the width of this character is greater than 1, move the index on by that much
					i += byteDelta;
					targetPotentialStringLength -= byteDelta; //Because the targetPotentialStringLength was measured in bytes, it should now be shorter by this much, in characters
					if (potentialStringIndex > 0)
					{
						//Continue an existing potential string
						potentialString[potentialStringIndex++] = chars[0];
						if (potentialStringIndex == targetPotentialStringLength)
						{
							//Completed a potential string
							//Write out any pending bytes as binhex
							//Subtract 3 to account for the length marker. There will always be one if there are any pending bytes
							int pendingBytesLength = i - potentialStringIndex - writeIndex - 3;
							if (potentialStringIndex < bytes.Length && pendingBytesLength > 0) //If the string length is the byte length, there are no bytes to write
							{
								writer.WriteStartElement("Hex");
								writer.WriteBinHex(bytes, writeIndex, pendingBytesLength);
								writer.WriteEndElement();
							}
							//Write out the string
							writer.WriteStartElement("String");
							if (forceLengthDescriptor) //This string started at the beginning of the byte array, but also had a length descriptor
								writer.WriteAttributeString("forceLengthDescriptor", XmlConvert.ToString(true));

							StringBuilder foundString = new StringBuilder(targetPotentialStringLength);
							if (chars[0] == 0) //This string is null terminated
							{
								writer.WriteAttributeString("nullTerminated", XmlConvert.ToString(true));
								targetPotentialStringLength--; //Subtract the terminator
							}

							foundString.Append(potentialString, 0, targetPotentialStringLength);

							writer.WriteString(foundString.ToString());
							writer.WriteEndElement();

							potentialStringIndex = 0;
							//Update write index
							writeIndex = i + 1;
						}
					}
					else
					{
						if (i == 0)
						{
							//Potentially the whole value is a string, so doesn't need a length marker
							targetPotentialStringLength = bytes.Length;
							potentialString = new char[targetPotentialStringLength];
							potentialString[0] = chars[0];
							potentialStringIndex = 1;
							forceLengthDescriptor = false;
						}
						else if (i > 3)
						{
							//Check for a length in the previous byte
							targetPotentialStringLength = BitConverter.ToInt32(bytes, i - 4);
							if (targetPotentialStringLength > 2 && targetPotentialStringLength <= bytes.Length - i) //If there aren't at least 3 chars, it isn't an interesting string anyway
							{
								//This is potentially a string, so start checking
								potentialString = new char[targetPotentialStringLength];
								potentialString[0] = chars[0];
								potentialStringIndex = 1;
								forceLengthDescriptor = i == 4; //If the string starts at pos 4, and 0-3 are the length marker, then force the length marker to be present.
							}
						}
					}
				}
				else //Not valid character
				{
					//Abandon any existing potential string
					potentialStringIndex = 0;
				}
			}

			//Write out any remaining bytes that weren't in the string
			if (writeIndex < bytes.Length)
			{
				writer.WriteStartElement("Hex");
				writer.WriteBinHex(bytes, writeIndex, bytes.Length - writeIndex);
				writer.WriteEndElement();
			}
		}

		private static bool UnicodeCategoryIsValid(UnicodeCategory unicodeCategory)
		{
			switch (unicodeCategory)
			{
				//Categories that indicate a valid string
				case UnicodeCategory.ClosePunctuation:
				case UnicodeCategory.ConnectorPunctuation:
				case UnicodeCategory.CurrencySymbol:
				case UnicodeCategory.DashPunctuation:
				case UnicodeCategory.DecimalDigitNumber:
				case UnicodeCategory.LowercaseLetter:
				case UnicodeCategory.MathSymbol:
				case UnicodeCategory.ModifierLetter:
				case UnicodeCategory.ModifierSymbol:
				case UnicodeCategory.OpenPunctuation:
				case UnicodeCategory.OtherPunctuation:
				case UnicodeCategory.SpaceSeparator:
				case UnicodeCategory.UppercaseLetter:
					return true;
				//Categories that indicate an invalid string
				case UnicodeCategory.Control:
				case UnicodeCategory.Format:
				case UnicodeCategory.EnclosingMark:
				case UnicodeCategory.FinalQuotePunctuation:
				case UnicodeCategory.InitialQuotePunctuation:
				case UnicodeCategory.LetterNumber:
				case UnicodeCategory.LineSeparator:
				case UnicodeCategory.NonSpacingMark:
				case UnicodeCategory.OtherLetter:
				case UnicodeCategory.OtherNotAssigned:
				case UnicodeCategory.OtherNumber:
				case UnicodeCategory.OtherSymbol:
				case UnicodeCategory.ParagraphSeparator:
				case UnicodeCategory.PrivateUse:
				case UnicodeCategory.SpacingCombiningMark:
				case UnicodeCategory.Surrogate:
				case UnicodeCategory.TitlecaseLetter:
					return false;
				default:
					System.Diagnostics.Debug.Fail("Unexpected unicode category");
					break;
			}
			return false;
		}

		public static Guid ReadGuid(Stream stream)
		{
			byte[] bytes = new byte[16];
			stream.Read(bytes, 0, 16);
			return new Guid(bytes);
		}

		public static int ReadInt(Stream stream)
		{
			byte[] bytes = new byte[4];
			stream.Read(bytes, 0, 4);
			return BitConverter.ToInt32(bytes, 0);
		}

		public static bool ReadBool(Stream stream)
		{
			byte[] bytes = new byte[1];
			stream.Read(bytes, 0, 1);
			return BitConverter.ToBoolean(bytes, 0);
		}

		/// <summary>
		/// Reads a null-terminated string starting from the length indicator
		/// </summary>
		public static string ReadString(Stream stream, bool nullTerminated)
		{
			int length = ReadInt(stream);

			if (length == 0)
				return String.Empty;

			byte[] bytes = new byte[length];
			stream.Read(bytes, 0, length);

			if (nullTerminated)
			{
				if (bytes[length - 1] != 0) //Check for null terminator
					Console.WriteLine("Unexpected string format! Check results carefully");
				else
					length -= 1; //Strip off the null terminator
			}

			return Encoding.UTF8.GetString(bytes, 0, length); //TODO: Check if the encoding is perhaps something freaky other than UTF8? 
		}

		public static string ReadEnum(Stream stream, string[] enumValues)
		{
			int value = ReadInt(stream);

			if (value < 0 || value >= enumValues.Length)
				return XmlConvert.ToString(value);

			return enumValues[value];
		}
	}
}
