using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;

namespace FooCfgToXml
{
	partial class Program
	{
		private static void ConvertCfgToXml(FileInfo fileInfo)
		{
			Dictionary<Guid, IValueInterpreter> valueInterpreters = Utilities.LoadAddins<Guid, IValueInterpreter>(delegate(IValueInterpreter valueInterpreter) { return valueInterpreter.SettingGuid; });

			FileInfo output = new FileInfo(String.Format("{0}.xml", fileInfo.FullName));
			Console.WriteLine("Converting \"{0}\"\n        to \"{1}\"\n", fileInfo.FullName, output.FullName);

			try
			{
				using (FileStream stream = fileInfo.OpenRead())
				{
					//Check guid signature at beginning
					if (CfgToXmlUtilities.ReadGuid(stream) != sFooConfigGuid)
					{
						Console.WriteLine("Aborting: unexpected config file format");
						return;
					}

					//Ignore the checksum
					stream.Seek(4, SeekOrigin.Current);

					//Read the number of settings blocks
					int numberOfSettingsBlocks = CfgToXmlUtilities.ReadInt(stream);

					//Start writing out the settings
					try
					{
						//Open the output file
						if (output.Exists)
						{
							Console.WriteLine(" Deleting existing \"{0}\"", output.Name);
							try
							{
								output.Delete();
							}
							catch (Exception e)
							{
								Console.WriteLine("Aborting: could not delete file: {0}", e.Message);
								return;
							}
						}

						XmlWriterSettings settings = new XmlWriterSettings();
						settings.Indent = true;
						settings.NewLineHandling = NewLineHandling.None;

						using(XmlWriter writer = XmlWriter.Create(output.FullName, settings))
						{
							writer.WriteStartDocument();
							writer.WriteStartElement("Config");
					
							for (int iSettingBlock = 0; iSettingBlock < numberOfSettingsBlocks; iSettingBlock++)
							{
								writer.WriteStartElement("Component");

								//Read the name of this block
								string name = CfgToXmlUtilities.ReadString(stream, true);
								writer.WriteAttributeString("name", name);

								//Ignore the null dword
								stream.Seek(4, SeekOrigin.Current);

								int blockLength = CfgToXmlUtilities.ReadInt(stream);
								long endOfBlockPosition = stream.Position + blockLength; //Add the block length to the current position to get the target end position

								//Read the settings
								while (stream.Position < endOfBlockPosition)
								{
									writer.WriteStartElement("Setting");

									Guid settingGuid = CfgToXmlUtilities.ReadGuid(stream);
									writer.WriteAttributeString("guid", settingGuid.ToString());

									int valueLength = CfgToXmlUtilities.ReadInt(stream);
									
									byte[] value = new byte[valueLength];
									stream.Read(value, 0, valueLength);

									//See if there is an interpreter for this value
									IValueInterpreter valueInterpreter;
									if (valueInterpreters.TryGetValue(settingGuid, out valueInterpreter))
									{
										writer.WriteAttributeString("interpreter", valueInterpreter.GetType().FullName);
										valueInterpreter.WriteValueToXml(writer, value);
									}
									else
									{
										CfgToXmlUtilities.WriteByteArrayToXml(writer, value);
									}
									
									writer.WriteEndElement();
								}

								writer.WriteEndElement();
							}
							writer.WriteEndDocument();
						}
					}
					catch (Exception e)
					{
						Console.WriteLine("Could not write to \"{0}\":\n{1}", output.Name, e.Message);
						return;
					}

					
				}
			}
			catch (Exception e)
			{
				Console.WriteLine("Could not read \"{0}\":\n{1}", fileInfo.Name, e.Message);
			}

			Console.WriteLine(" Successfully converted!");
		}
	}
}
