using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;

namespace FooCfgToXml
{
	/// <summary>
	/// This class contains helper methods for reading xml and writing to cfg binary.
	/// They are used internally, and can also be used by value interpreter extensions
	/// </summary>
	public static class XmlToCfgUtilities
	{
		/// <summary>
		/// The pair to WriteByteArrayToXml, this takes Xml written by that method
		/// and writes it to the filestream as a byte array.
		/// </summary>
		public static void WriteXmlToByteArray(Stream writer, XmlElement byteArrayXml)
		{
			if (byteArrayXml.ChildNodes.Count == 0)
				return; //No value

			bool beginsValue = true;

			using (XmlReader reader = new XmlTextReader(new StringReader(byteArrayXml.OuterXml)))
			{
				reader.MoveToContent();
				reader.ReadStartElement();
				do
				{
					switch (reader.Name)
					{
						case "Hex":
							int bytesRead;
							do
							{
								byte[] buffer = new byte[1024];
								bytesRead = reader.ReadElementContentAsBinHex(buffer, 0, buffer.Length);
								writer.Write(buffer, 0, bytesRead);
							} while (bytesRead > 0);
							break;
						case "String":
							bool includeLengthDescriptor;
							//Check if it has been specified
							string forceLengthDescriptor = reader.GetAttribute("forceLengthDescriptor");
							if (!String.IsNullOrEmpty(forceLengthDescriptor))
							{
								includeLengthDescriptor = XmlConvert.ToBoolean(forceLengthDescriptor);
							}
							else
							{
								//Not specified, so use default behaviour, 
								//which is always write the descriptor unless the string starts the value
								includeLengthDescriptor = !beginsValue;
							}

							bool terminate = false; //Not null terminated by default
							//Check if it has been specified
							string nullTerminated = reader.GetAttribute("nullTerminated");
							if (!String.IsNullOrEmpty(nullTerminated))
							{
								terminate = XmlConvert.ToBoolean(nullTerminated);
							}

							string str = reader.ReadElementContentAsString();
							WriteString(writer, str, includeLengthDescriptor, terminate);
							break;
						default:
							Console.WriteLine("Unknown setting value child element name ({0})! Check results carefully", reader.Name);
							reader.ReadInnerXml();
							break;
					}
					beginsValue = false;
				} while (reader.NodeType != XmlNodeType.EndElement);
			}
		}

		public static void WriteString(Stream writer, string str, bool writeLengthDescriptor, bool nullTerminated)
		{
			byte[] bytes = Encoding.UTF8.GetBytes(str);
			if (writeLengthDescriptor)
			{
				WriteInt(writer, bytes.Length + (nullTerminated ? 1 : 0));
			}
			writer.Write(bytes, 0, bytes.Length);
			if (nullTerminated)
			{
				writer.WriteByte(0);
			}
		}

		public static void WriteInt(Stream writer, int value)
		{
			byte[] bytes = BitConverter.GetBytes(value);
			writer.Write(bytes, 0, bytes.Length);
		}

		public static void WriteBool(Stream writer, bool value)
		{
			byte[] bytes = BitConverter.GetBytes(value);
			writer.Write(bytes, 0, bytes.Length);
		}

		public static void WriteGuid(Stream writer, Guid guid)
		{
			byte[] bytes = guid.ToByteArray();
			writer.Write(bytes, 0, bytes.Length);
		}

		public static void WriteEnum(Stream writer, string value, string[] enumValues)
		{
			int index = Array.IndexOf<string>(enumValues, value);
			if (index == -1)
				index = XmlConvert.ToInt32(value);

			WriteInt(writer, index);
		}

		/// <summary>
		/// Keep track of where the size of a section should be written.
		/// On disposal, fills in the amount that has been written to the stream since then
		/// as the size value (at the point in the stream it was constructed).
		/// </summary>
		public class SizeWriter : IDisposable
		{
			private long mSizePosition;
			private Stream mWriter;

			public SizeWriter(Stream writer)
			{
				mWriter = writer;
				mSizePosition = mWriter.Position;
				WriteInt(mWriter, 0); //Placeholder for size
			}

			public void Dispose()
			{
				//Go back and fill in the size
				long endPosition = mWriter.Position;
				mWriter.Seek(mSizePosition, SeekOrigin.Begin);
				WriteInt(mWriter, (int)(endPosition - mSizePosition - 4)); //Write the length of all blocks
				//Return to end of stream
				mWriter.Seek(endPosition, SeekOrigin.Begin);
			}
		}
	}
}
